#   Copyright 2021 Modelyst LLC
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

# Internal
from dbgen import Model
from dbgen.example.generators.analysis import analysis
from dbgen.example.generators.io import io
from dbgen.example.schema import all

################################################################################


def make_model() -> Model:
    # Initialize model
    m = Model("example")

    # Add objects and relations
    m.add(all)  # type: ignore

    # Add Generators
    io(m)
    analysis(m)
    return m
